#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko macrocycle builder
#

from collections import defaultdict
from operator import itemgetter

from .molsetup import Bond

# region
DEFAULT_MIN_RING_SIZE = 7
DEFAULT_MAX_RING_SIZE = 33
DEFAULT_DOUBLE_BOND_PENALTY = 50
DEFAULT_MAX_BREAKS = 4
# endregion


class FlexMacrocycle:
    """
    Attributes
    ----------
    _min_ring_size: int
    _max_ring_size: int
    _double_bond_penalty: float
    max_breaks: int
    setup:
    breakable_rings:
    """

    def __init__(
        self,
        min_ring_size: int = DEFAULT_MIN_RING_SIZE,
        max_ring_size: int = DEFAULT_MAX_RING_SIZE,
        double_bond_penalty: float = DEFAULT_DOUBLE_BOND_PENALTY,
        max_breaks: int = DEFAULT_MAX_BREAKS,
        allow_break_atype_A: bool = False,
    ):
        """
        Initialize macrocycle typer.

        Parameters
        ----------
        min_ring_size: int
            Minimum size of the ring, default is 7.
        max_ring_size: int
            Maximum size of the ring, default is 33.
        double_bond_penalty: float
        max_breaks: int
        allow_break_type_A: bool
            Allow breaking bonds involving atoms typed A, default is False.
        """
        self._min_ring_size = min_ring_size
        self._max_ring_size = max_ring_size
        # accept also double bonds (if nothing better is found)
        self._double_bond_penalty = double_bond_penalty
        self.max_breaks = max_breaks
        self.allow_break_atype_A = allow_break_atype_A

        self.setup = None
        self.breakable_rings = None

    def collect_rings(self, setup):
        """
        Gets non-aromatic rings of desired size and lists bonds that are part of unbreakable rings. Bonds belonging to
        rigid cycles can't be deleted or made rotatable even if they are part of a breakable ring.

        Parameters
        ----------
        setup: RDKitMoleculeSetup

        Returns
        -------
        breakable_rings: list
            A list of breakable ring ids
        bonds_in_rigid_cycles: set
             A set of the bonds in rigid cycles
        """
        breakable_rings = []
        rigid_rings = []
        for ring_id in list(
            setup.rings.keys()
        ):  # ring_id are the atom indices in each ring
            size = len(ring_id)
            if setup.rings[ring_id].is_aromatic:
                rigid_rings.append(ring_id)
            elif size < self._min_ring_size:
                rigid_rings.append(ring_id)
                # do not add rings > _max_ring_size to rigid_rings
                # because bonds in rigid rings will not be breakable
                # and these bonds may also belong to breakable rings
            elif size <= self._max_ring_size:
                breakable_rings.append(ring_id)

        bonds_in_rigid_cycles = set()
        for ring_atom_indices in rigid_rings:
            for bond in setup.get_bonds_in_ring(ring_atom_indices):
                bonds_in_rigid_cycles.add(bond)

        return breakable_rings, bonds_in_rigid_cycles

    def _score_bond(self, bond: tuple[int, int]) -> int:
        """
        Calculates a score for the likeliness that a bond will be broken.

        Parameters
        ----------
        bond: tuple[int, int]
            Input bond to score

        Returns
        -------
        score: int
            A score for the bond.
        """
        bond = Bond.get_bond_id(bond[0], bond[1])
        if not self.setup.bond_info[bond].rotatable:
            return -1
        atom_idx1, atom_idx2 = bond
        for i in (atom_idx1, atom_idx2):
            atype = self.setup.get_atom_type(i)
            is_allowed_A = self.allow_break_atype_A and atype == "A"
            if atype != "C" and not is_allowed_A:
                return -1
        # historically we returned a score <= 100, that was lower for triple
        # bonds, chiral atoms, conjugated bonds, and double bonds. This score
        # gets combined with the graph depth score in flexibility.py that
        # is lower when more consecutive torsions exist in a single "branch"
        # of the torsion tree. Any positive number can be returned here.
        return 100

    def get_breakable_bonds(self, bonds_in_rigid_rings):
        """
        Find breaking points for rings following the guidelines defined in [1].
        The optimal bond has the following properties:
        - does not involve a chiral atom
        - is not double/triple (?)
        - is between two carbons
        (preferably? we can now generate pseudoAtoms on the fly!)
        - is a bond present only in one ring

         [1] Forli, Botta, J. Chem. Inf. Model., 2007, 47 (4)
          DOI: 10.1021/ci700036j

        Parameters
        ----------
        bonds_in_rigid_rings: set[tuple]
            A set of bonds in rigid rings.

        Returns
        -------
        breakable: dict
            A dictionary of mapping breakable bonds to bond scores
        """
        breakable = {}
        for ring_atom_indices in self.breakable_rings:
            for bond in self.setup.get_bonds_in_ring(ring_atom_indices):
                score = self._score_bond(bond)
                if score > 0 and bond not in bonds_in_rigid_rings:
                    breakable[bond] = score
        return breakable

    def search_macrocycle(self, setup, delete_these_bonds=[]):
        """
        Search for macrocycles in the molecule

        Parameters
        ----------
        setup: RDKitMoleculeSetup
        delete_these_bonds: list

        Returns
        -------
        break_combo_data:
        bonds_in_rigid_rings:
        """
        self.setup = setup

        self.breakable_rings, bonds_in_rigid_rings = self.collect_rings(setup)
        if len(delete_these_bonds) == 0:
            breakable_bonds = self.get_breakable_bonds(bonds_in_rigid_rings)
        else:
            breakable_bonds = {}
            for bond in delete_these_bonds:
                bond = Bond.get_bond_id(bond[0], bond[1])
                breakable_bonds[bond] = self._score_bond(bond)
        break_combo_data = self.combinatorial_break_search(breakable_bonds)
        return break_combo_data, bonds_in_rigid_rings

    def combinatorial_break_search(self, breakable_bonds):
        """
        Enumerate all combinations of broken bonds. Once a bond is broken, it will break one or more rings. Subsequent
        bonds will be pulled from intact (unbroken) rings. The number of broken bonds may be variable.
        Returns only combinations of broken bonds that break the maximum number of broken bonds.

        Parameters
        ----------
        breakable_bonds: dict
            A dictionary mapping breakable bonds to the bond score.

        Returns
        -------
        break_combo_data: dict
            A dictionary containing information about possible bond break combinations
        """

        max_breaks = self.max_breaks
        break_combos = self._recursive_break(
            self.breakable_rings, max_breaks, breakable_bonds, set(), []
        )
        break_combos = list(break_combos)  # convert from set
        max_broken_bonds = 0
        output_break_combos = []  # found new max, discard prior data
        output_bond_scores = []
        output_unbroken_rings = []
        for broken_bonds in break_combos:
            n_broken_bonds = len(broken_bonds)
            bond_score = sum([breakable_bonds[bond] for bond in broken_bonds])
            if n_broken_bonds > max_broken_bonds:
                max_broken_bonds = n_broken_bonds
                output_break_combos = []  # found new max, discard prior data
                output_bond_scores = []
                output_unbroken_rings = []
            if n_broken_bonds == max_broken_bonds:
                output_break_combos.append(broken_bonds)
                output_bond_scores.append(bond_score)
                u = self.get_unbroken_rings(self.breakable_rings, broken_bonds)
                output_unbroken_rings.append(u)

        break_combo_data = {
            "bond_break_combos": output_break_combos,
            "bond_break_scores": output_bond_scores,
            "unbroken_rings": output_unbroken_rings,
        }
        return break_combo_data

    def _recursive_break(
        self, rings, max_breaks, breakable_bonds, output=set(), broken_bonds=[]
    ):
        """

        Parameters
        ----------
        rings: list
            List of rings to check
        max_breaks: int
            Maxumimum number of breaks allowed
        breakable_bonds:
        output: set
        broken_bonds:

        Returns
        -------
        output: set

        """
        if max_breaks == 0:
            return output
        unbroken_rings = self.get_unbroken_rings(rings, broken_bonds)
        atoms_in_broken_bonds = [
            atom_idx for bond in broken_bonds for atom_idx in bond
        ]
        for bond in breakable_bonds:
            if bond[0] in atoms_in_broken_bonds or bond[1] in atoms_in_broken_bonds:
                continue  # each atom can be in only one broken bond
            is_bond_in_ring = False
            for ring in unbroken_rings:
                if bond in self.setup.get_bonds_in_ring(ring):
                    is_bond_in_ring = True
                    break
            if is_bond_in_ring:
                current_broken_bonds = [(a, b) for (a, b) in broken_bonds + [bond]]
                num_unbroken_rings = len(
                    self.get_unbroken_rings(rings, current_broken_bonds)
                )
                data_row = tuple(sorted([(a, b) for (a, b) in current_broken_bonds]))
                output.add(data_row)
                if num_unbroken_rings > 0:
                    output = self._recursive_break(
                        rings,
                        max_breaks - 1,
                        breakable_bonds,
                        output,
                        current_broken_bonds,
                    )
        return output

    def get_unbroken_rings(self, rings, broken_bonds):
        """

        Parameters
        ----------
        rings:
        broken_bonds:

        Returns
        -------
        unbroken: list
            List of unbroken rings.
        """
        unbroken = []
        for ring in rings:
            is_unbroken = True
            for bond in broken_bonds:
                if bond in self.setup.get_bonds_in_ring(
                    ring
                ):  # consider precalculating bonds
                    is_unbroken = False
                    break  # pun intended
            if is_unbroken:
                unbroken.append(ring)
        return unbroken

    # region Deprecated

    def get_broken_rings(self, rings, broken_bonds):
        broken_rings = []
        for ring in rings:
            is_broken = False
            for bond in broken_bonds:
                if bond in self.setup.get_bonds_in_ring(
                    ring
                ):  # consider precalculating bonds
                    is_broken = True
                    break  # pun intended
            if is_broken:
                broken_rings.append(ring)
        return broken_rings

    def show_macrocycle_scores(self, setup):
        print("Warning: not showing macrocycle scores, check implementation.")
        return
        if setup is not None:
            print("\n==============[ MACROCYCLE SCORES ]================")
            bond_by_ring = defaultdict(list)

            for bond_id, data in list(setup.ring_bond_breakable.items()):
                ring_id = data["ring_id"]
                bond_by_ring[ring_id].append(bond_id)

            for ring_id, bonds in list(bond_by_ring.items()):
                data = []
                print(
                    "-----------[ ring id: %s | size: %2d ]-----------"
                    % (",".join([str(x) for x in ring_id]), len(ring_id))
                )

                for b in bonds:
                    score = setup.ring_bond_breakable[b]["score"]
                    data.append((b, score))

                data = sorted(data, key=itemgetter(1), reverse=True)

                for b_count, b in enumerate(data):
                    begin = b[0][0]
                    end = b[0][1]
                    # bond = self.mol.setup.get_bond(b[0][0], b[0][1])
                    # begin = bond.GetBeginAtomIdx()
                    # end = bond.GetEndAtomIdx()
                    info = (
                        b_count,
                        begin,
                        end,
                        b[1],
                        "#" * int(b[1] / 5),
                        "-" * int(20 - b[1] / 5),
                    )
                    print("[ %2d] Bond [%3d --%3d] s:%3d [%s%s]" % info)

    # endregion



from os import linesep as eol
import pathlib
import numpy as np

def get_gpf_string(center, size_angstrom, rec_fname, rec_types, lig_types, map_prefix=None, dielectric=-42, smooth=0.5, spacing=0.375, ff_param_fname=None):

    size_x, size_y, size_z = size_angstrom
    # the following guarantees an EVEN number of grid points
    npts_x = 2 * int(size_x / (2 * spacing))
    npts_y = 2 * int(size_y / (2 * spacing))
    npts_z = 2 * int(size_z / (2 * spacing))

    if map_prefix is None:
        map_prefix = pathlib.Path(rec_fname).with_suffix("").name

    gpf = ( "npts {gpf_npts_x:d} {gpf_npts_y:d} {gpf_npts_z:d}\n"
            "gridfld {map_prefix:s}.maps.fld\n"
            "spacing {gpf_spacing:1.3f}\n"
            "receptor_types {rec_types:s}\n"
            "ligand_types {lig_types_std:s}\n"
            "receptor {rec_fname:s}\n"
            "gridcenter {gpf_gridcenter_x:3.3f} {gpf_gridcenter_y:3.3f} {gpf_gridcenter_z:3.3f}\n"
            "smooth {gpf_smooth:3.3f}\n"
            "{mapfiles_std:s}"
            "elecmap {map_prefix:s}.e.map\n"
            "dsolvmap {map_prefix:s}.d.map\n"
            "dielectric {gpf_dielectric:3.3f}\n")

    if ff_param_fname is not None:
        gpf = "parameter_file %s\n" % ff_param_fname + gpf

    info = {}
    info["gpf_spacing"] = spacing
    info["gpf_smooth"] = smooth
    info["gpf_dielectric"] = dielectric
    info['map_prefix'] = map_prefix
    info['ff_param_fname'] = ff_param_fname
    info['rec_types'] = ' '.join(rec_types)
    info['rec_fname'] = rec_fname
    info['gpf_npts_x'] = npts_x
    info['gpf_npts_y'] = npts_y
    info['gpf_npts_z'] = npts_z
    info['gpf_gridcenter_x'] = center[0]
    info['gpf_gridcenter_y'] = center[1]
    info['gpf_gridcenter_z'] = center[2]

    mapfiles = ""
    for atype in lig_types:
        mapfiles+= "map %s.%s.map\n" % (map_prefix, atype)

    info['lig_types_std'] = " ".join(lig_types)
    info['mapfiles_std'] = mapfiles
    return gpf.format(**info), (npts_x, npts_y, npts_z)

def box_to_pdb_string(box_center, npts, spacing=0.375):
    """
        8 ______ 7
         /.    /|
      4 /_.___/3|
        | . X | |  <-- 9
        |5....|./6
        |.____|/
       1      2

    """

    step_x = int(npts[0] / 2.0) * spacing
    step_y = int(npts[1] / 2.0) * spacing
    step_z = int(npts[2] / 2.0) * spacing
    center_x, center_y, center_z = box_center
    corners = []
    corners.append([center_x - step_x, center_y - step_y, center_z - step_z] ) # 1
    corners.append([center_x + step_x, center_y - step_y, center_z - step_z] ) # 2
    corners.append([center_x + step_x, center_y + step_y, center_z - step_z] ) # 3
    corners.append([center_x - step_x, center_y + step_y, center_z - step_z] ) # 4
    corners.append([center_x - step_x, center_y - step_y, center_z + step_z] ) # 5
    corners.append([center_x + step_x, center_y - step_y, center_z + step_z] ) # 6
    corners.append([center_x + step_x, center_y + step_y, center_z + step_z] ) # 7
    corners.append([center_x - step_x, center_y + step_y, center_z + step_z] ) # 8

    count = 1
    res = "BOX"
    chain = "X"
    pdb_out = ""
    line = "ATOM  %5d %4s %3s %1s%4d    %8.3f%8.3f%8.3f  1.00 10.00          %1s" + eol
    for idx in range(len(corners)):
        x = corners[idx][0]
        y = corners[idx][1]
        z = corners[idx][2]
        pdb_out += line % (count, "Ne", res, chain, idx+1, x, y, z, "Ne")
        count += 1

    # center
    pdb_out += line % (count+1, "Xe", res, chain, idx+1, center_x, center_y, center_z, "Xe")

    pdb_out += "CONECT    1    2" + eol
    pdb_out += "CONECT    1    4" + eol
    pdb_out += "CONECT    1    5" + eol
    pdb_out += "CONECT    2    3" + eol
    pdb_out += "CONECT    2    6" + eol
    pdb_out += "CONECT    3    4" + eol
    pdb_out += "CONECT    3    7" + eol
    pdb_out += "CONECT    4    8" + eol
    pdb_out += "CONECT    5    6" + eol
    pdb_out += "CONECT    5    8" + eol
    pdb_out += "CONECT    6    7" + eol
    pdb_out += "CONECT    7    8" + eol
    return pdb_out

def is_point_outside_box(point, center, npts, spacing=0.375):

    # Autogrid always outputs an odd number of grid points, roundup if even.
    # An odd number of grid poins is an even number of voxels.
    halfvoxels = np.array(npts) / 2
    step = spacing * halfvoxels.astype(int)
    mincorner = center - step
    maxcorner = center + step
    x, y, z = point
    is_outside = False
    is_outside |= x >= maxcorner[0] or x <= mincorner[0]
    is_outside |= y >= maxcorner[1] or y <= mincorner[1]
    is_outside |= z >= maxcorner[2] or z <= mincorner[2]
    return is_outside

def calc_box(coord_array, padding):
    """Calulate gridbox around given coordinate array with given padding

    Args:
        coord_array (array-like): array of XYZ coordinates as strings or floats
        padding (string/float): padding (in angstroms) for box around given coordinates

    Returns:
        tuple: tuple of tuples with center coordinates (angstroms) and dimension sizes (angstroms) for box
    """    
    padding = float(padding)
    xa = [float(c[0]) for c in coord_array]
    ya = [float(c[1]) for c in coord_array]
    za = [float(c[2]) for c in coord_array]
    x_min = min(xa)
    x_max = max(xa)
    y_min = min(ya)
    y_max = max(ya)
    z_min = min(za)
    z_max = max(za)
    center_x = (x_min + x_max) / 2.0
    center_y = (y_min + y_max) / 2.0
    center_z = (z_min + z_max) / 2.0
    size_x = x_max - x_min + 2 * padding
    size_y = y_max - y_min + 2 * padding
    size_z = z_max - z_min + 2 * padding
    return (center_x, center_y, center_z), (size_x, size_y, size_z)

def box_to_vina_string(gridcenter, gridsizes):
    dims = ['x', 'y', 'z']
    centers_str = "\n".join([f"center_{d} = {gridcenter[i]:.3f}" for i, d in enumerate(dims)])
    sizes_str = "\n".join([f"size_{d} = {gridsizes[i]:.3f}" for i, d in enumerate(dims)])
    return centers_str + "\n" + sizes_str

boron_silicon_atompar  = "atom_par Si     4.10  0.200  35.8235  -0.00143  0.0  0.0  0  -1  -1  6" + eol
boron_silicon_atompar += "atom_par B      3.84  0.155  29.6478  -0.00152  0.0  0.0  0  -1  -1  0" + eol



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Reactive
#

from math import sqrt

class ReactiveAtomTyper:

    def __init__(self):

        self.ff = {
            "HD": {"rii": 2.00, "epsii": 0.020},
            "C": {"rii": 4.00, "epsii": 0.150},
            "A": {"rii": 4.00, "epsii": 0.150},
            "N": {"rii": 3.50, "epsii": 0.160},
            "NA": {"rii": 3.50, "epsii": 0.160},
            "OA": {"rii": 3.20, "epsii": 0.200},
            "OS": {"rii": 3.20, "epsii": 0.200},
            "F": {"rii": 3.09, "epsii": 0.080},
            "P": {"rii": 4.20, "epsii": 0.200},
            "SA": {"rii": 4.00, "epsii": 0.200},
            "S": {"rii": 4.00, "epsii": 0.200},
            "Cl": {"rii": 4.09, "epsii": 0.276},
            "CL": {"rii": 4.09, "epsii": 0.276},
            "Br": {"rii": 4.33, "epsii": 0.389},
            "BR": {"rii": 4.33, "epsii": 0.389},
            "I": {"rii": 4.72, "epsii": 0.550},
            "Si": {"rii": 4.10, "epsii": 0.200},
            "B": {"rii": 3.84, "epsii": 0.155},
            "W": {"rii": 0.00, "epsii": 0.000},
        }
        std_atypes = list(self.ff.keys())
        rt, r2s, r2o = self.enumerate_reactive_types(std_atypes)
        self.reactive_type = rt
        self.reactive_to_std_atype_mapping = r2s
        self.reactive_to_order = r2o

    @staticmethod
    def enumerate_reactive_types(atypes):
        reactive_type = {1: {}, 2: {}, 3: {}}
        reactive_to_std_atype_mapping = {}
        reactive_to_order = {}
        for order in (1, 2, 3):
            for atype in atypes:
                if len(atype) == 1:
                    new_atype = "%s%d" % (atype, order)
                else:
                    new_atype = "%s%d" % (atype[0], order + 3)
                    if new_atype in reactive_to_std_atype_mapping:
                        new_atype = "%s%d" % (atype[0], order + 6)
                        if new_atype in reactive_to_std_atype_mapping:
                            raise RuntimeError(
                                "ran out of numbers for reactive types :("
                            )
                reactive_to_std_atype_mapping[new_atype] = atype
                reactive_to_order[new_atype] = order
                reactive_type[order][atype] = new_atype
                ### # avoid atom type clashes with multiple reactive residues by
                ### # prefixing with the index of the residue, e.g. C3 -> 1C3.
                ### for i in range(8): # hopefully 8 reactive residues is sufficient
                ###     prefixed_new_atype = '%d%s' % ((i+1), new_atype)
                ###     reactive_to_std_atype_mapping[prefixed_new_atype] = atype
        return reactive_type, reactive_to_std_atype_mapping, reactive_to_order

    def get_scaled_parm(self, atype1, atype2):
        """generate scaled parameters for a pairwise interaction between two atoms involved in a
        reactive interaction

        Rij and epsij are calculated according to the AD force field:
            # - To obtain the Rij value for non H-bonding atoms, calculate the
            #        arithmetic mean of the Rii values for the two atom types.
            #        Rij = (Rii + Rjj) / 2
            #
            # - To obtain the epsij value for non H-bonding atoms, calculate the
            #        geometric mean of the epsii values for the two atom types.
            #        epsij = sqrt( epsii * epsjj )
        """

        atype1_org, _ = self.get_basetype_and_order(atype1)
        atype2_org, _ = self.get_basetype_and_order(atype2)
        atype1_rii = self.ff[atype1_org]["rii"]
        atype1_epsii = self.ff[atype1_org]["epsii"]
        atype2_rii = self.ff[atype2_org]["rii"]
        atype2_epsii = self.ff[atype2_org]["epsii"]
        atype1_rii = atype1_rii
        atype2_rii = atype2_rii
        rij = (atype1_rii + atype2_rii) / 2
        epsij = sqrt(atype1_epsii * atype2_epsii)
        return rij, epsij

    def get_reactive_atype(self, atype, reactive_order):
        """create or retrive new reactive atom type label name"""
        macrocycle_glue_types = ["CG%d" % i for i in range(9)]
        macrocycle_glue_types += ["G%d" % i for i in range(9)]
        if atype in macrocycle_glue_types:
            return None
        return self.reactive_type[reactive_order][atype]

    def get_basetype_and_order(self, atype):
        if len(atype) > 1:
            if atype[0].isdecimal():
                atype = atype[1:]  # reactive residues are prefixed with a digit
        if atype not in self.reactive_to_std_atype_mapping:
            return None, None
        basetype = self.reactive_to_std_atype_mapping[atype]
        order = self.reactive_to_order[atype]
        return basetype, order


reactive_typer = ReactiveAtomTyper()
get_reactive_atype = reactive_typer.get_reactive_atype


def atom_name_to_molsetup_index(monomer, atom_name):
    indices = []
    for atom in monomer.raw_rdkit_mol.GetAtoms():
        name = atom.GetPDBResidueInfo().GetName().strip()
        if name == atom_name:
            indices.append(atom.GetIdx())
    if len(indices) > 1:
        raise RuntimeError(f"multiple atoms matched query atom name {atom_name}")
    if len(indices) == 0:
        return None
    index = indices[0]
    index = monomer.mapidx_from_raw[index]
    inv = {j: i for i, j in monomer.molsetup_mapidx.items()}
    index = inv[index]
    return index


def assign_reactive_types_by_index(molsetup, reactive_atom_index, get_reactive_atype=get_reactive_atype):

    atypes = {atom.index: atom.atom_type for atom in molsetup.atoms}

    # type reactive atom
    original_type = molsetup.get_atom_type(reactive_atom_index)
    reactive_type = get_reactive_atype(original_type, reactive_order=1)
    atypes[reactive_atom_index] = reactive_type

    # type atoms 1 bond away from reactive atom
    for index1 in molsetup.atoms[reactive_atom_index].graph:
        if not molsetup.get_is_ignore(index1):
            original_type = molsetup.get_atom_type(index1)
            reactive_type = get_reactive_atype(original_type, reactive_order=2)
            atypes[index1] = reactive_type

        # type atoms 2 bonds away from reactive
        for index2 in molsetup.atoms[index1].graph:
            if index2 == reactive_atom_index:
                continue
            if not molsetup.get_is_ignore(index2):
                original_type = molsetup.get_atom_type(index2)
                reactive_type = get_reactive_atype(original_type, reactive_order=3)
                atypes[index2] = reactive_type
    return atypes


def assign_reactive_types(
    molsetup, smarts, smarts_idx, get_reactive_atype=get_reactive_atype
):

    atype_dicts = []
    for atom_indices in molsetup.find_pattern(smarts):

        reactive_atom_index = atom_indices[smarts_idx]
        atypes = assign_reactive_types_by_index(
            molsetup,
            reactive_atom_index,
            get_reactive_atype
        )



        atype_dicts.append(atypes)

    return atype_dicts


# enumerate atom type pair combinations for reactive docking configuration file


def get_reactive_config(
    types_1,
    types_2,
    eps12,
    r12,
    r13_scaling,
    r14_scaling,
    ignore=["HD", "F"],
    coeff_vdw=0.1662,
):
    """
    Args:
        types_1 (list): 1st set of atom types
        types_2 (list): 2nd set of atom types

    Returns:
        derivtypes (dict):
        modpairs (list):
    """

    # for reactive pair (1-2 interaction) use 13-7 potential
    n12 = 13
    m12 = 7

    # for 1-3 and 1-4 interactions stick to 12-6 potential
    n = 12
    m = 6

    derivtypes = {}
    for reactype in set(types_1).union(set(types_2)):
        basetype, order = reactive_typer.get_basetype_and_order(reactype)
        derivtypes.setdefault(basetype, [])
        derivtypes[basetype].append(reactype)

    scaling = {3: r13_scaling, 4: r14_scaling}

    modpairs = {}
    for t1 in set(types_1):
        basetype_1, order_1 = reactive_typer.get_basetype_and_order(t1)
        for t2 in set(types_2):
            basetype_2, order_2 = reactive_typer.get_basetype_and_order(t2)
            order = order_1 + order_2
            pair_id = tuple(sorted([t1, t2]))
            if order == 2:  # 1-2 interaction: these are the reactive atoms.
                modpairs[pair_id] = {"eps": eps12, "r_eq": r12, "n": n12, "m": m12}
            elif order == 3 or order == 4:
                if basetype_1 in ignore or basetype_2 in ignore:
                    rij = 0.01
                    epsij = 0.001
                else:
                    rij, epsij = reactive_typer.get_scaled_parm(t1, t2)
                    rij *= scaling[order]
                    epsij *= coeff_vdw
                modpairs[pair_id] = {"eps": epsij, "r_eq": rij, "n": n, "m": m}

    # pairs of types across sets that also happen within each set
    def enum_pairs(types):
        pairs = set()
        for i in range(len(types_1)):
            for j in range(i, len(types_1)):
                pair_id = tuple(sorted([types_1[i], types_1[j]]))
                pairs.add(pair_id)
        return pairs

    collisions = []
    collisions.extend([p for p in enum_pairs(types_1) if p in modpairs])
    collisions.extend([p for p in enum_pairs(types_2) if p in modpairs])
    collisions = set(collisions)

    return derivtypes, modpairs, collisions



#!/usr/bin/env python

from rdkit import Chem
from rdkit.Chem import rdDistGeom
from meeko import MoleculePreparation


def test():
    smiles = "C1CCCC1"
    mol = Chem.MolFromSmiles(smiles)
    mol = Chem.AddHs(mol)
    etkdgv3 = rdDistGeom.srETKDGv3()
    rdDistGeom.EmbedMolecule(mol, etkdgv3)
    mk_prep = MoleculePreparation()
    molsetup = mk_prep.prepare(mol)[0]
    nr_rot_bonds = 0
    for _, bond in molsetup.bond_info.items():
        nr_rot_bonds += int(bond.rotatable)
    assert nr_rot_bonds == 0



import json
import pathlib
import pytest

from meeko import Polymer
from meeko import PDBQTWriterLegacy
from meeko import MoleculePreparation
from meeko import ResidueChemTemplates
import meeko

from rdkit import Chem
import numpy as np


pkgdir = pathlib.Path(meeko.__file__).parents[1]

ahhy_example = pkgdir / "test/polymer_data/AHHY.pdb"
just_one_ALA_missing = (
    pkgdir / "test/polymer_data/just-one-ALA-missing-CB.pdb"
)
just_one_ALA = pkgdir / "test/polymer_data/just-one-ALA.pdb"
just_three_residues = pkgdir / "test/polymer_data/just-three-residues.pdb"
disulfide_bridge = pkgdir / "test/polymer_data/just_a_disulfide_bridge.pdb"
loop_with_disulfide = pkgdir / "test/polymer_data/loop_with_disulfide.pdb"
insertion_code = pkgdir / "test/polymer_data/1igy_B_82-83_has-icode.pdb"
non_sequential_res = pkgdir / "test/polymer_data/non-sequential-res.pdb"
has_altloc = pkgdir / "test/polymer_data/has-altloc.pdb"
has_lys = pkgdir / "test/polymer_data/has-lys.pdb"
has_lyn = pkgdir / "test/polymer_data/has-lyn.pdb"
has_lys_resname_lyn = pkgdir / "test/polymer_data/has-lys-resname-lyn.pdb"
disulfide_adjacent = pkgdir / "test/polymer_data/disulfide_bridge_in_adjacent_residues.pdb"


# TODO: add checks for untested polymer fields (e.g. input options not indicated here)

chem_templates = ResidueChemTemplates.create_from_defaults()
mk_prep = MoleculePreparation(
    merge_these_atom_types=["H"],
    charge_model="gasteiger",
    load_atom_params="ad4_types",
)


def check_charge(residue, expected_charge, tolerance=0.002):
    charge = 0
    for atom in residue.molsetup.atoms:
        if not atom.is_ignore:
            charge += atom.charge
    assert abs(charge - expected_charge) < tolerance


def run_padding_checks(residue):
    assert len(residue.molsetup_mapidx) == residue.rdkit_mol.GetNumAtoms()
    # check index mapping between padded and rea molecule
    for i, j in residue.molsetup_mapidx.items():
        padding_z = residue.padded_mol.GetAtomWithIdx(i).GetAtomicNum()
        real_z = residue.rdkit_mol.GetAtomWithIdx(j).GetAtomicNum()
        assert padding_z == real_z
    # check padding atoms are ignored
    for i in range(residue.padded_mol.GetNumAtoms()):
        if i not in residue.molsetup_mapidx:  # is padding atom
            assert residue.molsetup.atoms[i].is_ignore


def test_flexres_pdbqt():
    with open(loop_with_disulfide) as f:
        pdb_string = f.read()
    set_templates = {
        ":6": "CYX",
        ":17": "CYX",
    }  # TODO remove this to test use of bonds to set templates
    polymer = Polymer.from_pdb_string(
        pdb_string,
        chem_templates,
        mk_prep,
        set_templates,
        blunt_ends=[(":5", 0), (":18", 2)],
    )
    res11 = polymer.monomers[":11"]
    assert sum(res11.is_flexres_atom) == 0
    polymer.flexibilize_sidechain(":11", mk_prep)
    assert sum(res11.is_flexres_atom) == 9
    rigid, flex_dict = PDBQTWriterLegacy.write_from_polymer(polymer)
    nr_rigid_atoms = len(rigid.splitlines())
    assert nr_rigid_atoms == 124
    nr_flex_atoms = 0
    for line in flex_dict[":11"].splitlines():
        nr_flex_atoms += int(line.startswith("ATOM"))
    assert nr_flex_atoms == 9


def test_AHHY_all_static_residues():
    f = open(ahhy_example, "r")
    pdb_string = f.read()
    polymer = Polymer.from_pdb_string(
        pdb_string,
        chem_templates,
        mk_prep,
        blunt_ends=[("A:1", 0)],
    )
    # Asserts that the residues have been imported in a way that makes sense, and that all the
    # private functions we expect to have run as expected.
    assert len(polymer.monomers) == 4
    assert len(polymer.get_ignored_monomers()) == 0
    assert len(polymer.get_valid_monomers()) == 4
    assert polymer.monomers["A:1"].residue_template_key == "ALA"
    assert polymer.monomers["A:2"].residue_template_key == "HID"
    assert polymer.monomers["A:3"].residue_template_key == "HIE"
    assert polymer.monomers["A:4"].residue_template_key == "CTYR"

    check_charge(polymer.monomers["A:1"], 0.0)
    check_charge(polymer.monomers["A:2"], 0.0)
    check_charge(polymer.monomers["A:3"], 0.0)
    check_charge(polymer.monomers["A:4"], -1.0)

    pdbqt_strings = PDBQTWriterLegacy.write_string_from_polymer(polymer)
    rigid_part, movable_part = pdbqt_strings

    # remove newline chars because Windows/Unix differ
    rigid_part = "".join(rigid_part.splitlines())

    assert len(rigid_part) == 3555
    assert len(movable_part) == 0


def test_AHHY_padding():
    with open(ahhy_example, "r") as f:
        pdb_string = f.read()
    polymer = Polymer.from_pdb_string(
        pdb_string,
        chem_templates,
        mk_prep,
        blunt_ends=[("A:1", 0)],
    )
    assert len(polymer.monomers) == 4
    assert len(polymer.get_ignored_monomers()) == 0

    for residue_id in ["A:1", "A:2", "A:3", "A:4"]:
        residue = polymer.monomers[residue_id]
        run_padding_checks(residue)


def test_just_three_padded_mol():
    with open(just_three_residues, "r") as f:
        pdb_string = f.read()
    set_template = {":15": "NMET"}
    polymer = Polymer.from_pdb_string(
        pdb_string,
        chem_templates,
        mk_prep,
        set_template=set_template,
        blunt_ends=[(":17", 17)],
    )
    assert len(polymer.monomers) == 3
    assert len(polymer.get_ignored_monomers()) == 0
    assert len(polymer.get_valid_monomers()) == 3

    assert polymer.monomers[":15"].residue_template_key == "NMET"
    assert polymer.monomers[":16"].residue_template_key == "SER"
    assert polymer.monomers[":17"].residue_template_key == "LEU"
    check_charge(polymer.monomers[":15"], 1.0)
    check_charge(polymer.monomers[":16"], 0.0)
    check_charge(polymer.monomers[":17"], 0.0)

    for residue_id in [":15", ":16", ":17"]:
        residue = polymer.monomers[residue_id]
        run_padding_checks(residue)

    pdbqt_strings = PDBQTWriterLegacy.write_string_from_polymer(polymer)
    rigid_part, movable_part = pdbqt_strings
    # remove newline chars because Windows/Unix differ
    rigid_part = "".join(rigid_part.splitlines())
    assert len(rigid_part) == 2212
    assert len(movable_part) == 0


def test_AHHY_mutate_residues():
    # We want both histidines to be "HIP" and to delete the tyrosine
    set_template = {
        "A:2": "HIP",
        "A:3": "HIP",
    }
    delete_residues = ("A:4",)
    with open(ahhy_example, "r") as f:
        pdb_string = f.read()
    polymer = Polymer.from_pdb_string(
        pdb_string,
        chem_templates,
        mk_prep,
        residues_to_delete=delete_residues,
        set_template=set_template,
        blunt_ends=[("A:1", 0)],
    )
    assert len(polymer.monomers) == 3
    assert len(polymer.get_ignored_monomers()) == 0
    assert len(polymer.get_valid_monomers()) == 3

    assert polymer.monomers["A:1"].residue_template_key == "ALA"
    assert polymer.monomers["A:2"].residue_template_key == "HIP"
    assert polymer.monomers["A:3"].residue_template_key == "HIP"

    check_charge(polymer.monomers["A:1"], 0.0)
    check_charge(polymer.monomers["A:2"], 1.0)
    check_charge(polymer.monomers["A:3"], 1.0)

    pdbqt_strings = PDBQTWriterLegacy.write_string_from_polymer(polymer)
    rigid_part, movable_part = pdbqt_strings
    # remove newline chars because Windows/Unix differ
    rigid_part = "".join(rigid_part.splitlines())
    assert len(rigid_part) == 2528
    assert len(movable_part) == 0


def test_residue_missing_atoms():
    with open(just_one_ALA_missing, "r") as f:
        pdb_string = f.read()

    polymer = Polymer.from_pdb_string(
        pdb_string,
        chem_templates,
        mk_prep,
        allow_bad_res=True,
        blunt_ends=[("A:1", 0), ("A:1", 2)],
    )
    assert len(polymer.get_valid_monomers()) == 0
    assert len(polymer.monomers) == 1
    assert len(polymer.get_ignored_monomers()) == 1

    with pytest.raises(RuntimeError):
        polymer = Polymer.from_pdb_string(
            pdb_string,
            chem_templates,
            mk_prep,
            allow_bad_res=False,
            blunt_ends=[("A:1", 0), ("A:1", 2)],
        )
    return


def test_AHHY_mk_prep_and_export():
    with open(ahhy_example, "r") as f:
        pdb_text = f.read()
    mk_prep2 = MoleculePreparation(
        add_atom_types=[{"smarts": "[CH2,CH3]", "new_param": 42.0}]
    )
    polymer = Polymer.from_pdb_string(
        pdb_text,
        chem_templates,
        mk_prep2,
        blunt_ends=[("A:1", 0)],
    )
    ap, xyz = polymer.export_static_atom_params()
    # all parameters musthave same size
    assert len(set([len(values) for (key, values) in ap.items()])) == 1
    assert "new_param" in ap


def test_disulfides():
    with open(disulfide_bridge, "r") as f:
        pdb_text = f.read()
    # auto disulfide detection is enabled by default
    polymer_disulfide = Polymer.from_pdb_string(
        pdb_text,
        chem_templates,
        mk_prep,
        blunt_ends=[("B:22", 0), ("B:22", 2), ("B:95", 0), ("B:95", 2)],
    )
    # the disulfide bond is detected, and it expects two paddings,
    # but forcing CYS not CYX disables the padding, so error expected
    with pytest.raises(RuntimeError):
        polymer_thiols = Polymer.from_pdb_string(
            pdb_text,
            chem_templates,
            mk_prep,
            set_template={"B:22": "CYS"},
            blunt_ends=[("B:22", 0), ("B:22", 2), ("B:95", 0), ("B:95", 2)],
        )

    # remove bond and expect CYS between residues
    # currently, all bonds between a pair of residues will be removed
    polymer_thiols = Polymer.from_pdb_string(
        pdb_text,
        chem_templates,
        mk_prep,
        bonds_to_delete=[("B:22", "B:95")],
        blunt_ends=[("B:22", 0), ("B:22", 2), ("B:95", 0), ("B:95", 2)],
    )

    # check residue names
    assert polymer_disulfide.monomers["B:22"].residue_template_key == "CYX"
    assert polymer_disulfide.monomers["B:95"].residue_template_key == "CYX"
    assert polymer_thiols.monomers["B:22"].residue_template_key == "CYS"
    assert polymer_thiols.monomers["B:95"].residue_template_key == "CYS"


def test_insertion_code():
    with open(insertion_code, "r") as f:
        pdb_text = f.read()
    polymer = Polymer.from_pdb_string(
        pdb_text,
        chem_templates,
        mk_prep,
        blunt_ends=[("B:82", 0), ("B:83", 2)],
    )

    expected_res = set(("B:82", "B:82A", "B:82B", "B:82C", "B:83"))
    res = set(polymer.monomers)
    assert res == expected_res


def test_write_pdb_1igy():
    with open(insertion_code, "r") as f:
        pdb_text = f.read()
    polymer = Polymer.from_pdb_string(
        pdb_text,
        chem_templates,
        mk_prep,
        blunt_ends=[("B:82", 0), ("B:83", 2)],
    )
    pdbstr = polymer.to_pdb()

    # input 1igy has some hydrogens, here we are making sure
    # that the position of one of them didn't change
    expected = "  -7.232 -23.058 -15.763"
    found = False
    for line in pdbstr.splitlines():
        if line[30:54] == expected:
            found = True
            break
    assert found


def test_write_pdb_AHHY():
    with open(ahhy_example, "r") as f:
        pdb_text = f.read()
    polymer = Polymer.from_pdb_string(
        pdb_text,
        chem_templates,
        mk_prep,
        blunt_ends=[("A:1", 0)],
    )
    newpdbstr = polymer.to_pdb()
    # AHHy doesn't have hydrogens. If hydrogens get mangled xyz=(0, 0, 0) when
    # added by RDKit, we will probably not be able to match templates anymore.
    # and recreating the polymer from newpdbstr will very likely fail
    polymer = Polymer.from_pdb_string(
        newpdbstr,
        chem_templates,
        mk_prep,
        blunt_ends=[("A:1", 0)],
    )

def test_non_seq_res():
    """the residue atoms are interrupted (not in contiguous lines)
        which should cause the parser to throw an error. Here we
        check the an error is thrown.
    """
    with open(non_sequential_res, "r") as f:
        pdb_text = f.read()
    with pytest.raises(ValueError) as err_msg:
        polymer = Polymer.from_pdb_string(
            pdb_text,
            chem_templates,
            mk_prep,
        )
    assert str(err_msg.value).startswith("interrupted")

def test_altloc():
    with open(has_altloc, "r") as f:
        pdb_text = f.read()
    with pytest.raises(RuntimeError) as err_msg:
        polymer = Polymer.from_pdb_string(
            pdb_text,
            chem_templates,
            mk_prep,
        )
    assert "altloc" in str(err_msg.value).lower()

    polymer = Polymer.from_pdb_string(
        pdb_text,
        chem_templates,
        mk_prep,
        default_altloc="B",
    )
    res = polymer.monomers["A:264"]
    xyz = res.rdkit_mol.GetConformer().GetPositions()
    for atom in res.rdkit_mol.GetAtoms():
        index = atom.GetIdx()
        name = res.atom_names[index]
        if name == "OG":
            break
    assert abs(xyz[index][0] - 11.220) < 0.001

    polymer = Polymer.from_pdb_string(
        pdb_text,
        chem_templates,
        mk_prep,
        default_altloc="B",
        wanted_altloc={"A:264": "A"}
    )
    res = polymer.monomers["A:264"]
    xyz = res.rdkit_mol.GetConformer().GetPositions()
    for atom in res.rdkit_mol.GetAtoms():
        index = atom.GetIdx()
        name = res.atom_names[index]
        if name == "OG":
            break
    assert abs(xyz[index][0] - 12.346) < 0.001

def test_set_template_LYN():
    """the input is fully protonated NH3+"""
    with open(loop_with_disulfide) as f:
        pdb_string = f.read()
    polymer = Polymer.from_pdb_string(
        pdb_string,
        chem_templates,
        mk_prep,
        set_template={":16": "LYN"},
    )
    res16 = polymer.monomers[":16"]
    res17 = polymer.monomers[":17"]
    assert res17.residue_template_key == "CYX"
    assert res16.residue_template_key == "LYN"
    chrg16 = sum([a.charge for a in res16.molsetup.atoms if not a.is_ignore])
    assert abs(chrg16) < 1e-6

def test_weird_zero_coord():
    with open(has_lys) as f:
        pdbstr = f.read()
    polymer = Polymer.from_pdb_string(pdbstr, chem_templates, mk_prep)
    for _, res in polymer.monomers.items():
        positions = res.rdkit_mol.GetConformer().GetPositions()
        for atom in res.molsetup.atoms:
            # there was a bug in which the C-term CYS of has_lys would be
            # be assigned the erroneous CCYS template,and the extra oxygen
            # would get coordinates set to zero.
            assert np.min(np.sum(positions**2, 1)) > 1e-6

def test_auto_LYN():
    with open(has_lyn) as f:
        pdbstr = f.read()
    polymer = Polymer.from_pdb_string(pdbstr, chem_templates, mk_prep)
    assert polymer.monomers[":15"].residue_template_key == "LEU"
    assert polymer.monomers[":16"].residue_template_key == "LYN"
    assert polymer.monomers[":17"].residue_template_key == "CYX-"
    with open(has_lys) as f:
        pdbstr = f.read()
    polymer = Polymer.from_pdb_string(pdbstr, chem_templates, mk_prep)
    assert polymer.monomers[":16"].residue_template_key == "LYS"
    assert polymer.monomers[":17"].residue_template_key == "CYX-"
    polymer = Polymer.from_pdb_string(pdbstr, chem_templates, mk_prep, set_template={":16": "LYN"})
    assert polymer.monomers[":16"].residue_template_key == "LYN"
    assert polymer.monomers[":17"].residue_template_key == "CYX-"
    with open(has_lys_resname_lyn) as f:
        pdbstr = f.read()
    with pytest.raises(RuntimeError) as err_msg:
        polymer = Polymer.from_pdb_string(pdbstr, chem_templates, mk_prep)

def test_disulfide_adjacent():
    """ disulfide bridge in adjacent residues broke a version of the code
        that assumed only one bond between each pair of residues
    """
    with open(disulfide_adjacent, "r") as f:
        pdb_text = f.read()
    polymer = Polymer.from_pdb_string(
        pdb_text,
        chem_templates,
        mk_prep,
    )
